import numpy as np

import pyslammer as slam
import matplotlib.pyplot as plt

from pyslammer.decoupled_analysis import Decoupled
from pyslammer.constants import *
import math


class Coupled(Decoupled):
    def __init__(self,
                 ky: float or tuple[list[float], list[float]] or tuple[np.ndarray, np.ndarray] or callable,
                 a_in: list[float] or np.ndarray,
                 dt: float,
                 height: int or float,
                 vs_slope: int or float,
                 vs_base: int or float,
                 damp_ratio: float,
                 ref_strain: float,
                 scale_factor: float = 1,
                 soil_model: str = "linear_elastic",
                 si_units: bool = True,
                 lite: bool = False):
        super().__init__(ky, a_in, dt, height, vs_slope, vs_base, damp_ratio, ref_strain, scale_factor, soil_model,
                         si_units, lite)

        self.s1 = self.sdot1 = self.sdotdot1 = 0.0
        self.s2 = self.sdot2 = self.sdotdot2 = 0.0
        self.u1 = self.udot1 = self.udotdot1 = 0.0
        self.u2 = self.udot2 = self.udotdot2 = 0.0
        self.baseacc = self.basef = self.acc11 = self.acc22 = 0.0
        self.normalf1 = self.normalf2 = self.gameff1 = 0.0
        self.mx = self.mx1 = self.mmax = 0.0
        self.s = np.zeros(self.npts)
        self.u = np.zeros(self.npts)
        self.udotdot = np.zeros(self.npts)
        self.avgacc = np.zeros(self.npts)
        self.udot = np.zeros(self.npts)
        self.angle = 0
        self.COS = math.cos(self.angle * math.pi / 180.0)
        self.SIN = math.sin(self.angle * math.pi / 180.0)

        self.gCOS = self.g * self.COS
        self.gSIN = self.g * self.SIN

        self.beta = 0.25  # TODO: move to global constants
        self.gamma = 0.5  # TODO: move to global constants

        self.block_acc = np.zeros(self.npts)

        if type(self) is Coupled:
            self.run_sliding_analysis()
        self.ground_acc = a_in * self.g
        self.max_sliding_disp = self.block_disp[-1]


    def run_sliding_analysis(self):  # TODO: add ca to inputs
        if self.soil_model == "equivalent_linear":
            self.equivalent_linear()

        for i in range(1, self.npts + 1):
            self.dynamic_response(i)

        # calculate decoupled displacements
        for i in range(1, self.npts + 1):
            self.coupled_sliding(i)

        self.max_sliding_disp = self.block_disp[-1]
        # return self.max_sliding_disp
        self.block_disp = self.s

    def coupled_sliding(self, i):
        self.coupled_setupstate(i)
        # solve for x_resp, v_resp, a_resp at next time step
        self.solvu(i)
        self.udotdot[i - 1] = self.udotdot2

        # update sliding acceleration based on calc'd response
        self.c_slideacc(i)

        # check if sliding has started
        self.c_slidingcheck(i)
        self.block_acc[i-1] = self.acc22 - self.sdotdot2

        self.s[i - 1] = self.s2

    def coupled_setupstate(self, i):
        # set up state from previous time step
        if i == 1:
            self.u1 = self.udot1 = self.udotdot1 = self.s1 = self.sdot1 = self.sdotdot1 = self.normalf1 = 0.0
        else:
            self.u1 = self.u2
            self.udot1 = self.udot2
            self.udotdot1 = self.udotdot2
            self.s1 = self.s2
            self.sdot1 = self.sdot2
            self.sdotdot1 = self.sdotdot2
            self.normalf1 = self.normalf2

        # Set up acceleration loading. Normal force corrected for vertical component of a_in.
        self.normalf2 = self.mass * self.gCOS + self.mass * self.a_in[i - 1] * self.scale_factor * self.gSIN

        if i == 1:
            self.acc11 = 0.0
            self.acc22 = self.a_in[i - 1] * self.gCOS * self.scale_factor
        elif not self._slide:
            self.acc11 = self.a_in[i - 2] * self.gCOS * self.scale_factor
            self.acc22 = self.a_in[i - 1] * self.gCOS * self.scale_factor
        else:
            self.acc11 = self.gSIN - self.k_y(self.s[i - 2]) * self.normalf1 / self.mass
            self.acc22 = self.gSIN - self.k_y(self.s[i - 2]) * self.normalf2 / self.mass

    def solvu(self, i):
        khat = a = b = deltp = deltu = deltudot = d1 = 0.0

        delt = self.dt

        if self._slide:
            d1 = 1.0 - (self.L1 ** 2) / (self.M1 * self.mass)
        else:
            d1 = 1.0

        khat = (self._omega ** 2) + 2.0 * self._damp_tot * self._omega * self.gamma / (self.beta * delt) + d1 / (
                    self.beta * (delt ** 2))
        a = d1 / (self.beta * delt) + 2.0 * self._damp_tot * self._omega * self.gamma / self.beta
        b = d1 / (2.0 * self.beta) + delt * 2.0 * self._damp_tot * self._omega * (self.gamma / (2.0 * self.beta) - 1.0)

        if i == 1:
            deltp = -self.L1 / self.M1 * (self.acc22 - self.acc11)
            deltu = deltp / khat
            deltudot = self.gamma / (self.beta * delt) * deltu
            self.u2 = deltu
            self.udot2 = deltudot
            self.udotdot2 = (-(self.L1 / self.M1) * self.acc22 - 2.0 * self._damp_tot * self._omega * self.udot2 - (
                        self._omega ** 2) * self.u2) / d1
        else:
            deltp = -self.L1 / self.M1 * (self.acc22 - self.acc11) + a * self.udot1 + b * self.udotdot1
            deltu = deltp / khat
            deltudot = self.gamma / (self.beta * delt) * deltu - self.gamma / self.beta * self.udot1 + delt * (
                        1.0 - self.gamma / (2.0 * self.beta)) * self.udotdot1
            self.u2 = self.u1 + deltu
            self.udot2 = self.udot1 + deltudot
            self.udotdot2 = (-(self.L1 / self.M1) * self.acc22 - 2.0 * self._damp_tot * self._omega * self.udot2 - (
                        self._omega ** 2) * self.u2) / d1

        self.u[i - 1] = self.u2

    def c_slideacc(self, i):
        # update sliding acceleration based on calc'd response
        if self._slide:
            self.sdotdot2 = -self.a_in[i - 1] * self.gCOS * self.scale_factor - self.k_y(
                self.s[i - 2]) * self.normalf2 / self.mass - self.L1 * self.udotdot2 / self.mass + self.gSIN

        # calc. base force based on a_resp calc
        self.basef = -self.mass * self.a_in[
            i - 1] * self.gCOS * self.scale_factor - self.L1 * self.udotdot2 + self.mass * self.gSIN

        # If sliding is occurring, integrate sdotdot, using trapezoid rule, to get block_vel and block_disp.
        if self._slide:
            self.sdot2 = self.sdot1 + 0.5 * self.dt * (self.sdotdot2 + self.sdotdot1)
            self.s2 = self.s1 + 0.5 * self.dt * (self.sdot2 + self.sdot1)

    def c_slidingcheck(self, i):
        # check if sliding has started
        if not self._slide:
            if self.basef > self.k_y(self.s[i - 2]) * self.normalf2:
                self._slide = True
        elif self._slide:
            if self.sdot2 <= 0.0:
                self.slidestop(i)
                self._slide = False
                self.sdot2 = 0.0
                self.sdotdot2 = 0.0

    def slidestop(self, i):
        ddt = acc11 = acc22 = acc1b = delt = dd = 0.0
        khat = deltp = a = b = 0.0

        delt = self.dt

        # Time of end of sliding is taken as where block_vel=0 from previous analysis
        dd = -self.sdot1 / (self.sdot2 - self.sdot1)
        ddt = dd * delt
        acc11 = self.gSIN - self.k_y(self.s[i - 2]) * (self.gCOS + self.a_in[i - 1] * self.scale_factor * self.gSIN)
        acc1b = self.a_in[i - 2] * self.g * self.scale_factor + dd * (
                    self.a_in[i - 1] - self.a_in[i - 2]) * self.g * self.scale_factor
        acc22 = self.gSIN - self.k_y(self.s[i - 2]) * (self.gCOS + acc1b * self.SIN)

        # if dd=0, sliding has already stopped and skip this solution
        if dd == 0:
            return

        self.solvu(i)
        self.u1 = self.u2
        self.udot1 = self.udot2
        self.udotdot1 = self.udotdot2
        self.normalf2 = self.mass * self.gCOS + self.mass * acc1b * self.SIN
        self.sdotdot2 = -acc1b * self.COS - self.k_y(
            self.s[i - 2]) * self.normalf2 / self.mass - self.L1 * self.udotdot2 / self.mass + self.gSIN
        self.sdot2 = self.sdot1 + 0.5 * ddt * (self.sdotdot2 + self.sdotdot1)
        self.s2 = self.s1 + 0.5 * ddt * (self.sdot1 + self.sdot2)

        # Solve for non sliding response during remaining part of dt
        ddt = (1.0 - dd) * delt
        self._slide = False
        acc11 = acc22
        acc22 = self.a_in[i - 1] * self.gCOS * self.scale_factor

        khat = 1.0 + 2.0 * self._damp_tot * self._omega * self.gamma * ddt + (self._omega ** 2) * self.beta * (ddt ** 2)
        a = (1.0 - (self.L1 ** 2) / (self.mass * self.M1)) + 2.0 * self._damp_tot * self._omega * ddt * (
                    self.gamma - 1.0) + (self._omega ** 2) * (ddt ** 2) * (self.beta - 0.5)
        b = (self._omega ** 2) * ddt
        deltp = - self.L1 / self.M1 * (acc22 - acc11) + a * (self.udotdot1) - b * (self.udot1)
        self.udotdot2 = deltp / khat

        self.udot2 = self.udot1 + (1.0 - self.gamma) * ddt * (self.udotdot1) + self.gamma * ddt * (self.udotdot2)
        self.u2 = self.u1 + self.udot1 * ddt + (0.5 - self.beta) * (ddt ** 2) * (self.udotdot1) + self.beta * (
                    ddt ** 2) * (self.udotdot2)


equivalent_linear_testing = False
k_y_testing = False


def some_ky_func(disp):
    initial = 0.15
    minimum = 0.05
    factor = 0.005
    exponent = -1.5
    value = max(factor * (disp + minimum) ** exponent + 0.4 * disp, minimum)
    return min(initial, value)


if __name__ == "__main__":
    histories = slam.sample_ground_motions()
    ky_const = 0.15
    ky_interp = ([0.2, 0.3, 0.4, 0.5], [0.15, 0.14, 0.13, 0.12])
    ky_func = some_ky_func
    motion = histories["Chi-Chi_1999_TCU068-090"]
    # t_step = motion[0][1] - motion[0][0]
    # input_acc = motion[1] / 9.80665

    ca = slam.Coupled(ky=ky_func,
                      a_in=motion.accel,
                      dt=motion.dt,
                      height=50.0,
                      vs_slope=600.0,
                      vs_base=600.0,
                      damp_ratio=0.05,
                      ref_strain=0.0005,
                      scale_factor=1.0,
                      soil_model="equivalent_linear",
                      si_units=True,
                      lite=False)

    ca.run_sliding_analysis()
    print(ca.s[-1] * 100)


import numpy as np
import scipy.integrate as spint
import matplotlib.pyplot as plt

from pyslammer.sliding_block_analysis import SlidingBlockAnalysis

M_TO_CM = 100
G_EARTH = 9.80665 # Acceleration due to gravity (m/block_disp^2).


class RigidAnalysis(SlidingBlockAnalysis):
    """Rigid Block Analysis."""

    def __init__(self, a_in, dt, ky, method='jibson'):
        """
        Initialize rigid block analysis.
        Args:
            a_in (list): Ground acceleration (g).
            dt (float): Time step (s).
            ky (float): Critical acceleration (g).
            method (str, optional): Analysis method. Default is 'jibson'.
        """
        super().__init__()

        self.analysis_methods = {
            "jibson": self.jibson,
            "dgr": self.downslope_dgr,
            "gra": self.garcia_rivas_arnold,
        }
        self._npts = len(a_in)
        self.ground_acc = np.array(a_in) * G_EARTH
        self.dt = dt
        self.ky = ky*G_EARTH
        self.method = method

        analysis_function = self.analysis_methods.get(self.method)
        if analysis_function:
            analysis_function()
        else:
            print(f"Analysis type {self.method} is not supported.")
        pass

    def __str__(self):
        # if self.dt == -1.0:
        #     info = ('Record: {}\n'.format(self.name))
        # else:
        #     info = (
        #             'Rigid Block Analysis\n'
        #             +'Record  : {}\n'.format(self.name)
        #             +'PGA     : {:.3f} g\n'.format(self.pga)
        #             +'dt      : {:.3f} s\n'.format(self.dt)
        #             +'ky     : {:.3f} m/s^2\n'.format(self.ky)
        #             +'Disp    : {:.3f} m'.format(self.total_disp)
        #         )
        # return info
        #TODO: Re-implement
        return "Rigid Block Analysis"


    def jibson(self):
        """
        Calculate the downslope rigid block displacement, differential velocity, and acceleration using the Jibson method.
        """
        tol = 1e-5
        self.block_acc = np.zeros(len(self.ground_acc))
        self.sliding_vel = np.zeros(len(self.ground_acc))
        self.sliding_disp = np.zeros(len(self.ground_acc))
        # [previous, current]
        acc = [0, 0]
        vel = [0, 0]
        pos = [0, 0]

        for i in range(len(self.ground_acc)):
            gnd_acc_curr = self.ground_acc[i]
            if vel[1] < tol:
                if abs(gnd_acc_curr) > self.ky:
                    n = gnd_acc_curr / abs(gnd_acc_curr)
                else:
                    n = gnd_acc_curr / self.ky
            else:
                n = 1
            acc[1] = gnd_acc_curr - n * self.ky
            vel[1] = vel[0] + (self.dt / 2) * (acc[1] + acc[0])
            if vel[1] > 0:
                pos[1] = pos[0] + (self.dt / 2) * (vel[1] + vel[0])
            else:
                vel[1] = 0
                acc[1] = 0
            pos[0] = pos[1]
            vel[0] = vel[1]
            acc[0] = acc[1]
            self.sliding_disp[i] = pos[1]
            self.sliding_vel[i] = vel[1]
            self.block_acc[i] = gnd_acc_curr - acc[1]
        self.max_sliding_disp = self.sliding_disp[-1]

    def garcia_rivas_arnold(self):
        # for future implementation with velocity verlet
        pass

    def downslope_dgr(self):
        """
        Calculate the downslope rigid block displacement, differential velocity, and acceleration using the Jibson method.
        Args:
            ky (float, optional): Critical acceleration in multiples of g.
        Returns:
            None
        """
        if self.dt == -1.0:
            return
        else:
            self._clear_block_params()
            self.ky = k_y * G_EARTH
        time = np.arange(0, len(self.ground_acc) * self.dt, self.dt)
        block_sliding = False
        for i in range(len(self.gnd_acc)):
            if i == 0:
                self.block_acc.append(self.gnd_acc[i])
                self.block_vel.append(self.gnd_vel[i])
                continue
            tmp_block_vel = self.block_vel[i-1] + self.ky * self.dt
            if self.gnd_acc[i] > self.ky:
                block_sliding = True
            elif tmp_block_vel > self.gnd_vel[i]:
                block_sliding = False
            else:
                pass
            if block_sliding == True:
                self.block_vel.append(tmp_block_vel)
                self.block_acc.append(self.ky)
            else:
                self.block_acc.append(self.gnd_acc[i])
                self.block_vel.append(self.gnd_vel[i])
        self.block_vel = abs(self.gnd_vel - self.block_vel)
        self.block_disp = spint.cumulative_trapezoid(self.block_vel, time, initial=0)
        self.total_disp = self.block_disp[-1]

    def plot(self, acc: bool=True, vel: bool=True, disp: bool=True, gnd_motion: bool=False):
        """
        Plot the ground motion and the block response.
        Args:
            acc (bool, optional): Plot block acceleration.
            vel (bool, optional): Plot block differential velocity.
            disp (bool, optional): Plot block displacement.
            gnd_motion (bool, optional): Plot ground motion.
        Returns:
            None
        """
        num_plots = sum([acc, vel, disp])
        if self.dt == 1.0:
            return
        elif num_plots == 0:
            return
        elif len(self.block_acc) == 0:
            if gnd_motion:
                super().plot(acc, vel, disp, gnd_motion, called=False)
                return
            else:
                return
    # Perform rigid block downslope analysis via integration of relative velocity
    # between the block and the ground. Trapezoid integration is used to caluclate
    # ground velocity. Block velocity matches ground velocity until ground acceleration
    # exceeds the critical acceleration, at which point the block begins to _slide
    # accelerating at the critical acceleration. The block stops sliding when its velocity
    # exceeds ground velocity. Block displacement is then calculated by integrating
    # the relative velocity between the block and the ground.

        if time_history is None:
            return
        else:
            pass
        acc_crit = acc_crit * G_EARTH
        time = time_history[0][:]
        dt = time[1]-time[0]
        gnd_acc = time_history[1][:]
        gnd_vel = spint.cumulative_trapezoid(gnd_acc, time, initial=0)
        block_vel = np.copy(gnd_vel)
        block_acc = []
        block_sliding = False
        for i in range(len(gnd_acc)):
            if i == 0:
                continue
            tmp_block_vel = block_vel[i-1] + acc_crit*dt
            if gnd_acc[i] > acc_crit:
                block_sliding = True
            elif tmp_block_vel > gnd_vel[i]:
                block_sliding = False
            else:
                pass
            fig, ax = super().plot(acc, vel, disp, gnd_motion, called=True)
            fig.suptitle('Rigid Block Analysis\n{}'.format(self.name))
            remain_plots = num_plots
            if acc:
                if num_plots == 1:
                    acc = ax
                else:
                    i = num_plots - remain_plots
                    remain_plots -= 1
                    acc = ax[i]
                acc.plot(self.time, self.block_acc, label='Block Acceleration')
                acc.plot(self.time, [self.ky for i in range(len(self.time))], label='Critical Acceleration')
                acc.legend()
            if vel:
                if num_plots == 1:
                    vel = ax
                else:
                    j = num_plots - remain_plots
                    remain_plots -= 1
                    vel = ax[j]
                vel.plot(self.time, self.block_vel, label='Block Differential Velocity')
                vel.legend()
            if disp:
                if num_plots == 1:
                    disp = ax
                else:
                    k = num_plots - remain_plots
                    remain_plots -= 1
                    disp = ax[k]
                disp.plot(self.time, self.block_disp, label='Block Displacement')
                disp.legend()
            plt.show()


G_EARTH = 9.80665 # Acceleration due to gravity (m/s^2).

FT_TO_IN = 12.0
FT_TO_M = 0.3048
FT3_TO_IN3 = FT_TO_IN**3
CM_TO_FT = 0.032808399
CM_TO_IN = CM_TO_FT * FT_TO_IN
M_TO_CM = 100.0
M_TO_FT = 3.28084
M3_TO_CM3 = M_TO_CM**3
LBFT3_TO_KNM3 = 6.3659
KNM3_TO_LBFT3 = 1 / LBFT3_TO_KNM3

BETA = 0.25
GAMMA = 0.5



# Decoupled Block Analysis
#TODO: add docstrings
#TODO: add inherited variable values
#TODO: add "testing" features?
import numpy as np

import pyslammer as slam
import matplotlib.pyplot as plt

from pyslammer.sliding_block_analysis import SlidingBlockAnalysis
from pyslammer.constants import *
import math

def mod_damp_testing(effective_strain, ref_strain):
    g_over_gmax = strain_mod_update(effective_strain, ref_strain)
    new_damp = strain_damp_update(g_over_gmax, effective_strain, ref_strain)
    return g_over_gmax, new_damp

def strain_mod_update(effective_strain, ref_strain):
    return 1 / (1 + (effective_strain / ref_strain) ** 1) #TODO: move constants outside of function


def strain_damp_update(g_over_gmax, shear_strain, ref_strain):
    m1 = 1/math.pi
    m2 = shear_strain
    m3 = ref_strain
    m4 = shear_strain + ref_strain
    m5 = m4/m3
    m6 = m2**2 / m4
    masing_damping = m1*(4 * (m2-m3*math.log(m5)) / m6 - 2)


    new_damp = 0.62 * g_over_gmax**0.1 * masing_damping + 0.01 #TODO: move constants outside of function

    if equivalent_linear_testing:
        print(f"g_over_gmax: {g_over_gmax}")
        print(f"masing_damping: {masing_damping}")
        print(f"new_damp: {new_damp}")

    return new_damp


def impedance_damping(vs_base, vs_slope):
    # Adjust for base impedance. include_impedance_damping()?
    vs_ratio = vs_base / vs_slope
    damp_imp = 0.55016 * vs_ratio ** -0.9904 #TODO: move constants outside of function
    damp_imp = min(damp_imp, 0.2) #TODO: move constants outside of function
    return damp_imp


def constant_k_y(k_y):
    def _ky_func(disp):
        return k_y
    return _ky_func


def interpolated_k_y(k_y):
    def _ky_func(disp):
        disp_values, ky_values = k_y
        return np.interp(disp, disp_values, ky_values)
    if k_y_testing:
        for disp in [0, 10, 20, 30, 40, 50]:
            print(f"disp: {disp}, ky: {_ky_func(disp)}")
    return _ky_func


def assign_k_y(k_y):
    if isinstance(k_y, float):
        return constant_k_y(k_y)
    elif isinstance(k_y, tuple) and len(k_y) == 2:
        return interpolated_k_y(k_y)
    elif callable(k_y):
        return k_y
    else:
        val_error_msg = ("Invalid type for ky. Must be float, tuple, or callable."
                         "If tuple, must contain two equal-length lists or numpy arrays.")
        raise ValueError(val_error_msg)


class Decoupled(SlidingBlockAnalysis):
    # include allowed values for inputs, like soil_model
    def __init__(self,
                 ky: float or tuple[list[float], list[float]] or tuple[np.ndarray,np.ndarray] or callable,
                 a_in: list[float] or np.ndarray,
                 dt: float,
                 height: int or float,
                 vs_slope: int or float,
                 vs_base: int or float,
                 damp_ratio: float,
                 ref_strain: float,
                 scale_factor: float = 1,
                 soil_model: str = "linear_elastic",
                 si_units: bool = True,
                 lite: bool = False):
        super().__init__()
        self._npts = len(a_in)
        self.k_y = assign_k_y(ky)
        self.a_in = a_in #FIXME: inconsistent use of a_in with/without scale_factor
        self.dt = dt
        self.height = height
        self.vs_slope = vs_slope
        self.vs_base = vs_base
        self.damp_ratio = damp_ratio
        self.ref_strain = ref_strain
        self.SI_units = si_units
        self.soil_model = soil_model
        self.lite = lite

        self.scale_factor = scale_factor

        self.npts = len(a_in)
        self.g = G_EARTH*(si_units + (not si_units)*M_TO_FT)
        self.unit_weight = 20.0*(si_units + (not si_units)*KNM3_TO_LBFT3) #TODO: move constants outside of function
        self.rho = self.unit_weight / self.g  # DENSITY
        self.mass = self.unit_weight * height / self.g
        self.L1 = -2.0 * self.mass / math.pi * math.cos(math.pi)
        self.M1 = self.mass / 2.0
        self.max_shear_mod = self.rho * vs_slope**2

        self.HEA = np.zeros(self.npts)
        self.block_disp = np.zeros(self.npts)
        self.block_vel = np.zeros(self.npts)
        self.block_acc = np.zeros(self.npts)
        self.x_resp = np.zeros(self.npts)
        self.v_resp = np.zeros(self.npts)
        self.a_resp = np.zeros(self.npts)
        self.max_sliding_disp = 0.0

        # special variables that change during the analysis
        self._slide = False
        self._vs_slope = vs_slope
        self._omega = math.pi * vs_slope / (2.0 * height)
        self._damp_imp = impedance_damping(vs_base, vs_slope)
        self._damp_tot = damp_ratio + self._damp_imp

        if type(self) is Decoupled:
            self.run_sliding_analysis()
        self.ground_acc = a_in*self.g

    def run_sliding_analysis(self): #TODO: add ca to inputs

        if self.soil_model == "equivalent_linear":
            self.equivalent_linear()

        for i in range(1, self.npts + 1):
            self.dynamic_response(i)

        # calculate decoupled displacements
        for i in range(1, self.npts + 1):
            self.sliding(i)

        self.max_sliding_disp =  self.block_disp[-1]
        return self.max_sliding_disp

    def sliding(self, i): #TODO: refactor
        # variables for the previous and current time steps
        # prev and curr are equal for the first time step
        #TODO: consider just starting at i=2 and eliminating the logical statement in prev
        # alternatively, removing logical and letting it use the last value of the array...
        prev = i - 2 + (i == 1)
        curr = i - 1

        yield_acc = self.k_y(self.block_disp[prev]) * self.g
        excess_acc = yield_acc - self.HEA[prev]
        delta_hea = self.HEA[curr] - self.HEA[prev]

        if k_y_testing:
            if i%500 == 0:
                print(f"disp: {self.block_disp[prev]}, ky: {self.k_y(self.block_disp[prev])}")

        if not self._slide:
            self.block_acc[curr] = self.HEA[curr]
            self.block_vel[curr] = 0
            self.block_disp[curr] = self.block_disp[prev]
            if self.HEA[curr] > yield_acc:
                self._slide = True
        else:
            self.block_acc[curr] = yield_acc
            self.block_vel[curr] = (self.block_vel[prev]
                                    + (excess_acc - 0.5 * delta_hea) * self.dt)
            self.block_disp[curr] = (self.block_disp[prev]
                                     - self.block_vel[prev] * self.dt
                                     - 0.5 * (excess_acc + delta_hea / 6.0) * self.dt**2)
            if self.block_vel[curr] >= 0.0:
                self._slide = False

    def dynamic_response(self,i):
        prev = i - 2 + (i == 1)
        curr = i - 1
        # Newmark Beta Method constants
        beta = 0.25 #TODO: move outside of function (up to delta_a_in)
        gamma = 0.5 #TODO: move constants outside of function

        self._omega = math.pi * self._vs_slope / (2.0 * self.height) #TODO: move outside of function

        k_eff = (self._omega ** 2
                 + 2.0 * self._damp_tot * self._omega * gamma / (beta * self.dt)
                 + 1.0 / (beta * self.dt**2))
        a = (1.0 / (beta * self.dt)
             + 2.0 * self._damp_tot * self._omega * gamma / beta)
        b = (1.0 / (2.0 * beta)
             + 2.0 * self.dt * self._damp_tot * self._omega * (gamma / (2.0 * beta) - 1.0))

        delta_a_in = self.a_in[curr] - self.a_in[prev]
        delta_force = (- self.L1 / self.M1 * delta_a_in * self.g * self.scale_factor
                       + a * self.v_resp[prev]
                       + b * self.a_resp[prev])
        delta_x_resp = delta_force / k_eff
        delta_v_resp = (gamma / (beta * self.dt) * delta_x_resp
                    - gamma / beta * self.v_resp[prev]
                    + self.dt * (1.0 - gamma / (2.0 * beta)) * self.a_resp[prev])
        delta_a_resp = (1.0 / (beta * (self.dt * self.dt)) * delta_x_resp
                       - 1.0 / (beta * self.dt) * self.v_resp[prev]
                       - 0.5 / beta * self.a_resp[prev])

        self.x_resp[curr] = self.x_resp[prev] + delta_x_resp
        self.v_resp[curr] = self.v_resp[prev] + delta_v_resp
        self.a_resp[curr] = self.a_resp[prev] + delta_a_resp

        self.HEA[curr] = self.a_in[curr]*self.g + self.L1 / self.mass * self.a_resp[curr]



    def equivalent_linear(self):
        tol = 0.05 #TODO: move constants outside of function
        max_iterations = 100 #TODO: move constants outside of function
        rel_delta_mod = 1
        rel_delta_damp = 1
        shear_mod = self.max_shear_mod
        damp_ratio = self.damp_ratio
        count = 0
        while rel_delta_damp > tol or rel_delta_mod > tol: #TODO: confirm whether number of iterations and order of operations matches SLAMMER
            for i in range(1, self.npts + 1):
                self.dynamic_response(i)
            peak_disp = max(abs(self.x_resp))
            effective_strain = 0.65 * 1.57 * peak_disp / self.height #TODO: move constants outside of function
            g_over_gmax = strain_mod_update(effective_strain, self.ref_strain)
            new_mod = g_over_gmax * self.max_shear_mod
            new_damp = strain_damp_update(g_over_gmax, effective_strain, self.ref_strain)

            if equivalent_linear_testing:
                print(f"iteration: {count}")
                print(f"effective_strain: {effective_strain}")
                print(f"_vs_slope: {self._vs_slope}")
                print(f"_damp_tot: {self._damp_tot}")


            self._vs_slope = math.sqrt(new_mod / self.rho)
            self._damp_imp = impedance_damping(self.vs_base, self._vs_slope)
            self._damp_tot = new_damp + self._damp_imp

            rel_delta_mod = abs((new_mod - shear_mod) / shear_mod)
            rel_delta_damp = abs((new_damp - damp_ratio) / damp_ratio)

            damp_ratio = new_damp
            shear_mod = new_mod

            count += 1
            if count > max_iterations:
                print("Warning: Maximum iterations reached. Equivalent linear procedure did not converge.")



mrd_testing = False
equivalent_linear_testing = False
k_y_testing = False

def some_ky_func(disp):
    initial = 0.15
    minimum = 0.05
    factor = 0.005
    exponent = -1.5
    value = max(factor*(disp+minimum)**exponent + 0.4*disp, minimum)
    return min(initial,value)

if __name__ == "__main__":
    if mrd_testing:
        strains = np.linspace(0.0001, 0.1, 1000)
        mod_reduction = []
        damping = []
        for strain in strains:
            mod_reduction.append(strain_mod_update(strain, 0.0005))
            damping.append(strain_damp_update(strain_mod_update(strain, 0.0005), strain, 0.0005))
        darendelli = [mod_damp_testing(strain, 0.0005) for strain in strains]
        plt.close("all")
        # plt.plot(mod_reduction, damping)
        plt.semilogx(strains, mod_reduction)
        plt.semilogx(strains, damping)
        plt.show()
    else:
        histories = slam.sample_ground_motions()
        ky_const = 0.15
        ky_interp = ([0.2, 0.3, 0.4, 0.5], [0.15, 0.14, 0.13, 0.12])
        ky_func = some_ky_func
        motion = histories["Chi-Chi_1999_TCU068-090"]
        t_step = motion[0][1] - motion[0][0]
        input_acc = motion[1] / 9.80665

        da = slam.Decoupled(ky=ky_const,
                            a_in=input_acc,
                            dt=t_step,
                            height=50.0,
                            vs_slope=600.0,
                            vs_base=600.0,
                            damp_ratio=0.05,
                            ref_strain=0.0005,
                            scale_factor=1.0,
                            soil_model="equivalent_linear",
                            si_units=True,
                            lite=False)

        da.run_sliding_analysis()

        print(da.block_disp[-1]*100)


"""
analytical_sliding_block.py

This module contains the implementation of the analytical model for a sliding block.
It defines the necessary symbols and parameters for the model, and sets up the global variables.

The module imports the necessary libraries, including numpy, sympy, pandas, and matplotlib.

Functions:
    set_globals(): Sets the global variables for the model.
    show_solution_approach(): Displays the definitions of the input acceleration, velocity, and displacement.
    create_harmonic_input_files(freq, resolution, cycles=10): Creates input files for a harmonic motion with a given frequency, resolution, and number of cycles.
    find_harmonic_solution(freq_val, ky_val, grav = 9.81, plot=True): Finds the solution for a harmonic motion with a given frequency and stiffness, optionally plots the solution.
    find_t1(a_in, ky, vals): Finds the time at which the input acceleration equals the stiffness.
    find_t2(v_in, vb, vals, freq_val): Finds the time at which the input velocity equals the block velocity.
    find_displacement(v_in, vb, t1, t2, vals): Finds the displacement of the block.
    apply_find_harmonic_solution(row, plot = False): Applies the find_harmonic_solution function to a row of data.
    harmonic_solutions(harmonic_combinations, save=False, plot=False): Finds the solutions for a set of harmonic combinations, optionally saves the solutions to a file.
    harmonic_solution_plot(a_in, v_in, vb, displacement,time,t1_def_val,t2_val, ky, vals, save=False): Plots the solution for a harmonic motion.

Global variables:
    time, freq, g, ky: Symbols for the analytical solution.
    a_in, v_inmax, v_in, x_in: Function expressions for the model.

Usage:
    Import this module to use the analytical model for a sliding block.
    Call set_globals() to set the global variables.
"""


import pickle as pkl
import numpy as np
import sympy as sym
import pandas as pd
import matplotlib.pyplot as plt

sym.init_printing()

# for analytical model
frequencies = np.array([0.5,1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
yield_accelerations = np.array([0.1,0.2,0.4,0.6,0.8,1.0])
combinations = np.array(np.meshgrid(frequencies, yield_accelerations)).T.reshape(-1, 2)

# for numerical models
sample_resolution = np.array([50, 100, 200, 250, 500, 750, 1000]) # samples per second

def set_globals():
    """
    Set global variables and defines symbols for the analytical solution.

    This function sets the values of global variables `time`, `freq`, `g`, `ky`, 
    and defines symbols for the analytical solution. 
    It also defines the function expressions for `a_in`, `v_inmax`, `v_in`, and `x_in`.

    Parameters:
        None

    Returns:
        None
    """
    global time, freq, g, ky, a_in, v_inmax, v_in, x_in
    # Define symbols for analytical solution
    # Define parameters
    time, freq, g, ky = sym.symbols('t, f, g, ky', real=True, positive=True)
    # time, freq, g, ky, t1, t2 = sym.symbols('t, f, g, ky, t_1, t_2', real=True, positive=True)
    # display(time, freq, g, ky, t1, t2)

    # Define the function expressions
    a_in = sym.sin(freq*2*sym.pi*time)
    v_inmax = g/2/sym.pi/freq
    v_in = g*sym.integrate(a_in, time) + v_inmax
    x_in = sym.integrate(v_in, time)

set_globals()

def show_solution_approach():
    """
    Display the solution approach for the sliding block problem.
    
    This function displays the equations defining the input acceleration, velocity, and displacement
    for the sliding block problem.
    
    Returns:
        None
    """
    display("The input acceleration (in units of g) is defined by:")
    display(sym.Eq(sym.symbols('\ddot{x_resp}_{in}(t)/g'), a_in))
    display("The input velocity (in m/block_disp) is defined by:")
    display(sym.Eq(sym.symbols('\dot{x_resp}_{in}(t)'), v_in))
    display("The input displacement (in m) is defined by:")
    display(sym.Eq(sym.symbols('u_{in}(t)'), x_in))
    return None

def create_harmonic_input_files(freq, resolution, cycles=10):
    """
    Create harmonic input files.

    Parameters:
    - freq (float): The frequency of the sine wave in Hz.
    - resolution (float): The resolution of the time series in samples per second.
    - cycles (int): The number of cycles of the sine wave.

    Returns:
    None
    """
    duration = cycles/freq
    t = np.linspace(0, duration, int(duration*resolution*cycles))
    a = np.sin(np.pi*2*freq*t)
    data = np.column_stack((t, a))
    np.savetxt(f'sample_ground_motions/sine_{freq}_Hz_{resolution*freq}_sps.csv', data, delimiter=',', header=f'# Time Series: {freq} Hz sine wave \n # Time (block_disp),Acceleration (g)')
    return None


def find_harmonic_solution(freq_val, ky_val, grav=9.81, plot=True):
    """
    Finds the harmonic solution for a sliding block system.

    Parameters:
        freq_val (float): The frequency value.
        ky_val (float): The ky value.
        grav (float, optional): The gravitational constant. Defaults to 9.81.
        plot (bool, optional): Whether to plot the harmonic solution. Defaults to True.

    Returns:
        float: The total displacement of the block.
    """
    vals = [(g, grav), (ky, ky_val), (freq, freq_val)]

    t1 = find_t1(a_in, ky, vals)
    vb = g * ky * (time - t1) + v_in.subs(time, t1)
    t2 = find_t2(v_in, vb, vals, freq_val)

    dx_block = find_displacement(v_in, vb, t1, t2, vals)
    total_disp = float(dx_block.subs(time, t2).subs(vals))
    if plot:
        harmonic_solution_plot(a_in, v_in, vb, dx_block, time, t1, t2, ky, vals)
    return total_disp

def find_t1(a_in, ky, vals):
    """
    Find the value of t1 by solving the equation a_in - ky = 0.

    Parameters:
    a_in (symbolic expression): The input acceleration function.
    ky (float): The value of ky.
    vals (dict): A dictionary containing variable substitutions for the equation.

    Returns:
    float: The value of t1 after substituting the variable values.
    """
    t1 = sym.solve(a_in-ky, time)[1]
    return t1.subs(vals)

def find_t2(v_in, vb, vals, freq_val):
    """
    Find the value of t2 by solving the equation v_in - vb = 0.

    Parameters:
    v_in (symbolic expression): The input velocity function.
    vb (symbolic expressionn): The block velocity function.
    vals (dict): A dictionary containing variable substitutions for the equation.
    freq_val (float): The frequency value. This guides the solver to find the correct root.

    Returns:
    float: The value of t2, representing the time at which the input velocity equals the block velocity.
    """
    equation = sym.Eq(vb, v_in).subs(vals)
    t2 = sym.nsolve(equation, time, (0.25/freq_val, 1/freq_val), solver='bisect')
    return t2

def find_displacement(v_in, vb, t1, t2, vals):
    """
    Calculate the displacement of a sliding block over time.

    Parameters:
    v_in (symbolic expression): Velocity of the block before time t1.
    vb (symbolic expression): Velocity of the block during time t1 to t2.
    t1 (float): Start time of the block'block_disp sliding motion.
    t2 (float): End time of the block'block_disp sliding motion.
    vals (dict): Dictionary of values to substitute into the expressions.

    Returns:
    displacement (symbolic expression): Piecewise function representing the displacement of the block over time.
    """
    x_in = sym.integrate(v_in, time)
    x_block = sym.integrate(vb, time) + x_in.subs(time,t1) - sym.integrate(vb, time).subs(time,t1)
    total_displacement = x_in.subs(vals).subs(time,t2) - x_block.subs(vals).subs(time,t2)
    displacement = sym.Piecewise((0, time < t1), (total_displacement, time > t2),(x_in - x_block, True))
    return displacement

def apply_find_harmonic_solution(row, plot=False):
    """
    Apply the find_harmonic_solution function to a given row of data.

    Parameters:
    - row: A pandas DataFrame row containing the 'Frequency (Hz)' and 'ky (g)' values.
    - plot: A boolean indicating whether to plot the results.

    Returns:
    - The result of the find_harmonic_solution function.

    """
    return find_harmonic_solution(row['Frequency (Hz)'], row['ky (g)'], plot=plot)


def harmonic_solutions(harmonic_combinations, save=False, plot=False):
    """
    Calculate the harmonic solutions for given combinations of frequency and ky values.
    
    Args:
        harmonic_combinations (list): A list of tuples containing the frequency (Hz) and ky (g) values.
        save (bool, optional): Whether to save the harmonic solutions to a CSV file. Defaults to False.
        plot (bool, optional): Whether to plot the harmonic solutions. Defaults to False.
    
    Returns:
        pandas.DataFrame: A DataFrame containing the harmonic solutions with columns for frequency, ky, and displacement.
    """
    harmonic_solutions = pd.DataFrame(harmonic_combinations, columns=['Frequency (Hz)', 'ky (g)'])
    harmonic_solutions['Displacement (m)'] = harmonic_solutions.apply(lambda row: apply_find_harmonic_solution(row, plot=plot), axis=1)
    if save:
        harmonic_solutions.to_csv('common/harmonic_solutions.csv', index=False)
    return harmonic_solutions

def harmonic_solution_plot(a_in, v_in, vb, displacement, time, t1_def_val, t2_val, ky, vals, save=False):
    """
    Plot the harmonic solution of a sliding block system.

    Parameters:
    a_in (sympy.Expr): The acceleration input.
    v_in (sympy.Expr): The velocity input.
    vb (sympy.Expr): The velocity during the block phase.
    displacement (sympy.Expr): The displacement of the block.
    time (sympy.Symbol): The time variable.
    t1_def_val (float): The value of time at which the block phase starts.
    t2_val (float): The value of time at which the block phase ends.
    ky (sympy.Symbol): The stiffness coefficient.
    vals (dict): A dictionary of symbol-value pairs used for substitution.
    save (bool, optional): Whether to save the plot as an image file. Defaults to False.

    Returns:
    None
    """
    
    ablock = sym.Piecewise((a_in, time < t1_def_val), (a_in, time > t2_val),(ky, True))
    aplot = sym.lambdify(time, a_in.subs(vals), 'numpy')
    abplot = sym.lambdify(time, ablock.subs(vals), 'numpy')
    vplot = sym.lambdify(time, v_in.subs(vals), 'numpy')
    vbplot = sym.lambdify(time, vb.subs(vals), 'numpy')
    
    displacementplot = sym.lambdify(time, displacement.subs(vals), 'numpy')
    plt.close('all')
    fig, axs = plt.subplots(3, 1, sharex=True, figsize=(6, 8))
    time_vals = np.linspace(0, 1/vals[2][1], 1000)
    t1_index = np.argmin(np.abs(time_vals - t1_def_val))
    t2_index = np.argmin(np.abs(time_vals - t2_val))

    axs[0].plot(time_vals, ky.subs(vals)*np.ones(len(time_vals)), 'k--', linewidth=0.5)
    axs[0].plot(time_vals, aplot(time_vals))
    axs[0].plot(time_vals, abplot(time_vals))
    axs[0].legend(['$ky$', 'Input', 'Block'])
    axs[0].set_title('Acceleration')
    axs[0].set_ylabel('Acceleration, g')

    axs[1].plot(time_vals, vplot(time_vals))
    axs[1].plot(time_vals[t1_index:t2_index], vbplot(time_vals)[t1_index:t2_index])
    axs[1].fill_between(time_vals[t1_index:t2_index], vbplot(time_vals)[t1_index:t2_index], vplot(time_vals)[t1_index:t2_index], color='gray', alpha=0.5)
    axs[1].set_title('Velocity')
    axs[1].set_ylabel('Velocity, m/block_disp')

    tmax = max(time_vals)
    dmax = displacementplot(tmax)
    axs[2].plot(time_vals, displacementplot(time_vals),'tab:orange')
    axs[2].text(0.95, 0.5, f'Total Displacement: {round(float(dmax),3)} m', horizontalalignment='right', verticalalignment='center', transform=axs[2].transAxes)
    axs[2].set_title('Displacement')
    axs[2].set_xlabel('Time, block_disp')
    axs[2].set_ylabel('Displacement, m')

    if save:
        plt.savefig(f'harmonic_solution_freq-{vals[2][1]}_ky-{vals[1][1]}.png')
    return None
    
if __name__ == '__main__':
    find_harmonic_solution(1, 0.6, grav = 9.81, plot=True)
    plt.show()


import ipywidgets as widgets
import pyslammer.utilities as util
import numpy as np

def setup_widgets():
    """
    Sets up the interactive widgets for the demo.

    Returns:
    gm_widget (ipywidgets.Dropdown): The dropdown widget for selecting the ground motion.
    ky_widget (ipywidgets.BoundedFloatText): The bounded float text widget for setting the input PGA.
    analysis_widget (ipywidgets.ToggleButtons): The toggle buttons widget for selecting the analysis type.
    start_widget (ipywidgets.Button): The button widget for starting the analysis.
    """
    # Setting the yield acceleration
    ky_widget = widgets.BoundedFloatText(
        value=0.35,
        min=0,
        max=5.0,
        step=0.01,
        description='ky (g):',
        disabled=False
    )
    # Selecting the ground motion
    selections = [(name, time_history) for name, time_history in util.sample_ground_motions().items()]
    gm_widget = widgets.Dropdown(
        options = selections,
        value=selections[0][1],
        equals=np.array_equal,
        description='Select ground motion:'
    )

    analysis_widget = widgets.ToggleButtons(
        options=['Rigid', 'Decoupled', 'Coupled'],
        description='Analysis type:',
        disabled=False,
        button_style='', # 'success', 'info', 'warning', 'danger' or ''
        tooltips=[],
    #     icons=['check'] * 3
    )

    # Start analysis
    start_widget = widgets.Button(
        description='Run Analysis',
        disabled=False,
        button_style='warning', # 'success', 'info', 'warning', 'danger' or ''
        tooltip='Run Analysis',
        icon='refresh', # (FontAwesome names without the `fa-` prefix)
        value = None
    )

    return gm_widget, ky_widget, analysis_widget, start_widget


from pyslammer.sliding_block_analysis import *
from pyslammer.rigid_analysis import *
from pyslammer.decoupled_analysis import *
from pyslammer.coupled_analysis import *
from pyslammer.utilities import *

# import pyslammer.analytical_sliding_block as asb
import pyslammer.demo_widgets as dw

__version__ = "0.1.19"

import os




import numpy as np
import pyslammer.constants as constants
import scipy.integrate as spint
from scipy.fft import rfft, rfftfreq
import matplotlib.pyplot as plt

#TODO: bring this into utilities.py

class GroundMotion():
    """Ground Motion Record."""
    def __init__(self, accel: np.ndarray or list, dt: float, name: str='None'):
        """
        Creates a ground motion record object.
        Args:
            acceleration (np.ndarray or list): Ground motion acceleration record in g
            dt (float): Time step of the record (s).
            name (str): Name of the record.
        Creates:
            _gnd_motion (np.ndarray): Copy of original ground motion record.
            _is_scaled (bool): True if the record is scaled.
            _is_inverted (bool): True if the record is inverted.
            name (str): Name of the record.
            time (np.ndarray): Time vector of the record (s).
            dt (float): Time step of the record (s).
        """
        self.accel = np.array(accel)
        self.dt = dt
        self.name = name
        self._is_inverted = False
        self._is_scaled = False
        self._npts = len(accel)
        self.pga = max(abs(accel))

        # FFT
        x = rfft(accel)[1::]
        freqs = rfftfreq(self._npts, dt)[1::]
        x_real = np.real(x)
        x_imag = np.imag(x)
        c = np.sqrt(x_real ** 2 + x_imag ** 2)

        self.mean_period = sum(c ** 2 / freqs) / sum(c ** 2)

    def __str__(self):
        return f"Ground Motion: {self.name}, PGA: {self.pga:.3f} g, dt: {self.dt:.3f} s, npts: {self._npts}"
    
    def _calc_gnd_params(self):
        """
        Semi-private method to initialize and recalculate ground motion parameters.
        Args:
            None
        Creates:
            gnd_acc (np.ndarray): Ground acceleration (m/s^2).
            gnd_vel (np.ndarray): Ground velocity (m/s).
            gnd_disp (np.ndarray): Ground displacement (m).
            pga (float): Peak ground acceleration in multiples of g.
        """
        self.gnd_acc = self._gnd_motion[1][:] * constants.G_EARTH
        self.gnd_vel = spint.cumulative_trapezoid(self.gnd_acc, self.time, initial=0)
        self.gnd_disp = spint.cumulative_trapezoid(self.gnd_vel, self.time, initial=0)
        self.pga = max(abs(self.gnd_acc)) / constants.G_EARTH
    
    def scale(self, pga: float=False, scale_factor: float=False):
        """
        Scale the ground motion using desired method. Does nothing if more than one method is selected.
        Args:
            pga (float, optional): Desired peak ground acceleration in g.
            scale_factor (float, optional): Desired scale factor.
        Returns: 
            None
        """
        if self.dt == -1.0:
            return
        else:
            pass
        if (pga and scale_factor):
            return
        else:
            if self._is_scaled:
                self.unscale()
            else:
                pass
        if pga:
            scale_factor = pga / self.pga
            self.gnd_acc *= scale_factor
            self.gnd_vel *= scale_factor
            self.gnd_disp *= scale_factor
            self.pga = pga
        elif scale_factor:
            self.gnd_acc *= scale_factor
            self.gnd_vel *= scale_factor
            self.gnd_disp *= scale_factor
            self.pga *= scale_factor
        else:
            return        
        self._is_scaled = True
        self.name = self.name + '_SCALED'

    def unscale(self):
        """
        Unscales the ground motion.
        Args:
            None
        Returns:
            None
        """
        if self.dt == -1.0:
            return
        else:
            pass
        self._calc_gnd_params()
        self.name = self.name.replace('_SCALED', '')

    def invert(self):
        """
        Invert the ground motion.
        Args:
            None
        Returns:
            None
        """
        if self.dt == -1.0:
            return
        else:
            pass
        if self._is_inverted:
            self.uninvert()
            return
        else:
            self.gnd_acc *= -1
            self.gnd_vel *= -1
            self.gnd_disp *= -1
        self._is_inverted = True
        self.name = self.name + '_INVERTED'

    def uninvert(self):
        """
        Uninverts the ground motion.
        Args:
            None
        Returns:
            None
        """
        if self.dt == -1.0:
            return  
        else:
            pass
        if self._is_inverted:
            self.gnd_acc *= -1
            self.gnd_vel *= -1
            self.gnd_disp *= -1
            self._is_inverted = False
            self.name = self.name.replace('_INVERTED', '')
        else:
            return
    
    def plot(self, acc: bool=True, vel: bool=True, disp: bool=True, enable: bool=True, called: bool=False):
        """
        Plots desired ground motion parameters.
        Args:
            acc (bool, optional): Plot acceleration.
            vel (bool, optional): Plot velocity.
            disp (bool, optional): Plot displacement.
            enable (bool, optional): Enable plotting of ground parameters. Used if called from a RigidBlock object.
            called (bool, optional): True if called from a RigidBlock object.
        Returns:
            fig, ax (plt.figure, plt.axis): Figure and axis objects if called from a RigidBlock object.
        """
        if self.dt == -1.0:
            return
        else:
            pass
        num_plots = sum([acc, vel, disp])
        remain_plots = num_plots
        if num_plots == 0:
            return
        elif num_plots == 1:
            fig, ax = plt.subplots(num=self.name)
            ax.set_xlabel('Time (s)')
        else:
            fig, ax = plt.subplots(num_plots, 1, num=self.name)
            ax[-1].set_xlabel('Time (s)')
        fig.suptitle('Ground Motion\n{}'.format(self.name))
        if enable:
            pass
        else:
            return fig, ax
        if acc:
            if num_plots == 1:
                acc = ax
            else:
                i = num_plots - remain_plots
                remain_plots -= 1
                acc = ax[i]
            acc.plot(self.time, self.gnd_acc, label='Ground Acceleration')
            acc.set_ylabel('Acceleration (m/s^2)')
            acc.set_title('Ground Acceleration')
            acc.legend()
        if vel:
            if num_plots == 1:
                vel = ax
            else:
                j = num_plots - remain_plots
                remain_plots -= 1
                vel = ax[j]
            vel.plot(self.time, self.gnd_vel, label='Ground Velocity')
            vel.set_ylabel('Velocity (m/s)')
            vel.set_title('Ground Velocity')
            vel.legend()
        if disp:
            if num_plots == 1:
                disp = ax
            else:
                k = num_plots - remain_plots
                remain_plots -= 1
                disp = ax[k]
            disp.plot(self.time, self.gnd_disp, label='Ground Displacement')
            disp.set_ylabel('Displacement (m)')
            disp.set_title('Ground Displacement')
            disp.legend()
        if called:
            return fig, ax
        else:
            plt.show()


# This file is in the public domain.
import pyslammer as slam
from pyslammer.sliding_block_analysis import SlidingBlockAnalysis

import math

class DynamicResp(SlidingBlockAnalysis):
    def __init__(self):
        super().__init__()
        # main function parameters
        self.uwgt = 0.0
        self.height = 0.0
        self.vs = 0.0
        self.damp = 0.0
        self.damp1 = 0.0
        self.dt = 0.0
        self.scal = 0.0
        self.g = 0.0
        self.vr = 0.0
        self.vs1 = 0.0
        self.mmax = 0.0
        self.dv2 = True
        self.eqvlnr = False

        # main function variables
        self.Mtot = 0.0
        self.M = 0.0
        self.L = 0.0
        self.omega = 0.0
        self.beta = 0.0
        self.gamma = 0.0
        self.angle = 0.0
        self.qq = 0
        self.nmu = 0
        self.npts = 0

        self.rho = 0.0
        self.delt = 0.0
        self.dampf = 0.0
        self.damps = 0.0
        self.damps_prev = 0.0
        self.j = 0

        # _slide=0 no sliding, _slide=1 sliding
        # variable that end in 1 are for previous time step
        # variable that end in 2 are for current time step

        self.slide = False

        self.mx = 0.0
        self.mx1 = 0.0
        self.gameff1 = 0.0
        self.gamref = 0.0
        self.n = 0.0
        self.o = 0.0
        self.acc1 = 0.0
        self.acc2 = 0.0
        self.u1 = 0.0
        self.udot1 = 0.0
        self.udotdot1 = 0.0
        self.s = []
        self.u = []
        self.udot = []
        self.udotdot = []
        self.disp = []
        self.mu = []
        self.avgacc = []

        self.ain = []






import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate as spint

from pyslammer.constants import G_EARTH


class SlidingBlockAnalysis:

    def __init__(self):
        self.method = None
        self.ky = None
        self.time = None

        self.ground_acc = None
        self.ground_vel = None
        self.ground_disp = None

        self.block_acc = None
        self.block_vel = None
        self.block_disp = None

        self.sliding_vel = None
        self.sliding_disp = None
        self.max_sliding_disp = None
        self._npts = None
        pass

    def tbd(self):
        time = np.arange(0, self._npts * self.dt, self.dt)
        if self.ground_vel is None:
            self.ground_vel = spint.cumulative_trapezoid(self.ground_acc, time, initial=0)
            self.ground_disp = spint.cumulative_trapezoid(self.ground_vel, time, initial=0)
        if self.sliding_vel is None:
            self.sliding_vel = self.ground_vel - self.block_vel
        if self.sliding_disp is None:
            self.sliding_disp = self.block_disp# - self.ground_disp
        pass

    def sliding_block_plot(self, sliding_vel_mode=True, fig=None):
        self.tbd()
        bclr = "k"
        gclr = "tab:blue"
        kyclr = "k"
        if fig is None:  # gAcc,gVel,bVel,bDisp,t,ky):
            fig, axs = plt.subplots(3, 1, sharex=True)
            fig.set_size_inches(10, 6)
        else:
            axs = fig.get_axes()
        time = np.arange(0, self._npts * self.dt, self.dt)

        # axs[0].plot(time, self.ky * np.ones(len(time))/G_EARTH, label='Yield Acc.', linestyle='--',
        #             color=kyclr, linewidth=0.5)
        axs[0].plot(time, self.ground_acc/G_EARTH, label='Ground Acc.', color=gclr)
        axs[0].plot(time, self.block_acc/G_EARTH, label='Block Acc.', color=bclr)
        axs[0].set_ylabel('Acc. (g)')
        axs[0].set_xlim([time[0], time[-1]])

        if sliding_vel_mode:
            axs[1].plot(time, self.sliding_vel, label='Sliding Vel.', color=bclr)
        else:
            axs[1].plot(time, self.ground_vel, label='Ground Vel.', color=gclr)
            axs[1].plot(time, self.block_vel, label='Block Vel.', color=bclr)
        axs[1].set_ylabel('Vel. (m/s)')

        axs[2].plot(time, self.sliding_disp, label='Sliding Disp.', color=bclr)
        axs[2].set_ylabel('Disp. (m)')
        for i in range(len(axs)):
            axs[i].set_xlabel("Time (s)")
            axs[i].grid(which='both')
            # Place the legend outside the plot area
            axs[i].legend(loc='upper left', bbox_to_anchor=(1, 1))
        fig.tight_layout()
        fig.canvas.toolbar_position = 'top'
        return fig, axs






import importlib.resources as pkg_resources
from pathlib import Path
import numpy as np
import csv
from pyslammer.record import GroundMotion

G_EARTH = 9.80665

__all__ = ['csv_time_hist', 'sample_ground_motions']


def sample_ground_motions():
    sgms = {}

    # Get the path to the sample_ground_motions folder
    folder_path = pkg_resources.files('pyslammer') / "sample_ground_motions"

    # Iterate over all files in the folder
    for file_path in folder_path.glob("*.csv"):
        # Add the file name to the list
        sgms[file_path.name[:-4]] = GroundMotion(*csv_time_hist(file_path))

    return sgms


def csv_time_hist(filename: str):
    """
    Read a CSV file containing time history acceleration data and return a 1D numpy array and a timestep

    Returns:
        a_in: A 1D numpy array containing time history data.
        dt: The timestep of the data.
    """
    file = open(filename, 'r')
    if file is None:
        return None
    else:
        pass
    reader = csv.reader(file)
    time = []
    accel = []
    for row in reader:
        if '#' in row[0]:
            continue
        else:
            pass
        if len(row) == 2:
            time.append(float((row[0])))
            accel.append(float((row[1])))
        else:
            accel.append(float((row[0])))
    dt = time[1] - time[0]
    accel = np.array(accel)
    return accel, dt